module Main where
    
import Control.Parallel.Strategies

main :: IO()
main = print(r)



{-
    Make a move on the board
    Input: board, tar_cor, side
    Return: board
-}
makeAMove :: [[Int]] -> (Int,Int) -> Int -> [[Int ]]
makeAMove board tar_cor side = makeAMoveHelper board tar_cor 0 side

makeAMoveHelper :: [[Int]] -> (Int, Int) -> Int -> Int -> [[Int]]
makeAMoveHelper (x:xs) (tar_x, tar_y) curr_x side
    | tar_x == curr_x = (makeAMoveRow x tar_y side) : xs
    | otherwise = x:(makeAMoveHelper xs (tar_x, tar_y) (curr_x+1) side)

makeAMoveRow :: [Int] -> Int -> Int -> [Int]
makeAMoveRow row index side = makeAMoveRowHelper row index 0 side

makeAMoveRowHelper :: [Int] -> Int -> Int -> Int -> [Int]
makeAMoveRowHelper (x:xs) tar_index curr_index side
    | tar_index == curr_index = side : xs 
    | otherwise = x:(makeAMoveRowHelper xs tar_index (curr_index+1) side)




{-
    Score a board for one board
    Input: board 
    Return: score (int)
-}

scoreBoard :: [[Int]] -> Int
scoreBoard board = (scoreBoardOneSide board 1) - (scoreBoardOneSide board (-1))

scoreBoardOneSide :: [[Int]] -> Int -> Int
scoreBoardOneSide board side = (scoreBoardOneDirection board side) + (scoreBoardOneDirection (flipBoard board) side)

scoreOneRow :: [Int] -> Int -> Int
scoreOneRow row side = scoreOneRowHelper row side 0

scoreHelper :: Int -> Int
scoreHelper num = case num of 
    2 -> 100
    3 -> 1000
    4 -> 10000
    5 -> 100000
    _ -> 0

scoreOneRowHelper :: [Int] -> Int -> Int -> Int
scoreOneRowHelper [] side num = scoreHelper num
scoreOneRowHelper row@(x:xs) side num
        | x == side = scoreOneRowHelper xs side (num + 1)
        | otherwise = (scoreHelper num) + (scoreOneRowHelper xs side 0)

scoreBoardOneDirection :: [[Int]] -> Int -> Int
scoreBoardOneDirection board side = sum [scoreOneRow row side | row <- board]

flipBoard :: [[Int]] -> [[Int]]
flipBoard matrix 
    | null matrix = [[] | _ <- [1 .. 9]]
    | otherwise = let (x:xs) = matrix in
        [a:b | (a,b) <- (zip x (flipBoard xs))]


{-
    Get all possible moves on the board
    Input: board
    Return: moves [(int,int)]
-}

getAllMoves :: [[Int]] -> [(Int, Int)]
getAllMoves board = getAllMovesHelper board 0

getAllMovesHelper :: [[Int]] -> Int -> [(Int, Int)]
getAllMovesHelper board row_index
    | null board = []
    | otherwise = let (x:xs) = board in
        (getAllMovesOneRow x row_index) ++ (getAllMovesHelper xs (row_index + 1))

getAllMovesOneRow :: [Int] -> Int -> [(Int, Int)]
getAllMovesOneRow row row_index = getAllMovesOneRowHelper row row_index 0

getAllMovesOneRowHelper :: [Int] -> Int -> Int -> [(Int, Int)]
getAllMovesOneRowHelper row row_index curr_index
    | null row = []
    | otherwise = let (x:xs) = row in 
        do
            if x == 0
                then (row_index, curr_index):(getAllMovesOneRowHelper xs row_index (curr_index + 1))
            else getAllMovesOneRowHelper xs row_index (curr_index + 1)


{-
    AI functions
-}

initializeBestScore :: Int -> Int
initializeBestScore side 
    | side == 1 = -100000
    | otherwise = 100000

switchSide :: Int -> Int
switchSide side = -side

chooseBetterScore :: Int -> (Int,(Int, Int)) -> (Int,(Int, Int)) -> (Int,(Int, Int))
chooseBetterScore side (score1, move1) (score2, move2) = 
    do 
        if (side == 1 && score1 > score2) || (side == -1 && score1 < score2)
            then (score1, move1)
        else
            (score2, move2)

getBestMoveHelper :: [[Int]] -> Int -> Int -> Int -> (Int, Int) -> (Int,(Int, Int))
getBestMoveHelper board side depth curr_depth move
    | curr_depth == depth = (scoreBoard board, move)
    | otherwise = chooseBestMove allResults side
    where possibleMoves = getAllMoves board
          allResults = parMap rdeepseq (parallelHelper side board depth curr_depth) possibleMoves


chooseBestMove :: [(Int,(Int, Int))] -> Int -> (Int,(Int, Int))
chooseBestMove [] side | side == 1 = (-100000, (10000,10000))
                       | otherwise = (100000, (10000,10000))
chooseBestMove results@(x:xs) side = chooseBetterScore side nextR x
                                    where nextR = chooseBestMove xs side

parallelHelper :: Int -> [[Int]] -> Int -> Int -> (Int,Int) -> (Int, (Int, Int))
parallelHelper side board depth curr_depth move = getBestMoveHelper movedBoard (switchSide side) depth (curr_depth + 1) move
                                                where movedBoard = makeAMove board move side

getBestMove :: [[Int]] -> Int -> Int -> (Int, Int)
getBestMove board side depth = snd (getBestMoveHelper board (switchSide side) depth 0 (10000, 10000))


-- ================ testing -- ================


board1 = [
        [0,1,1,0,0,0,0,0,0],
        [0,1,0,0,0,0,0,0,0],
        [0,0,0,1,1,1,0,0,0],
        [0,0,0,0,0,0,0,0,0],
        [0,0,0,-1,-1,0,0,0,0],
        [0,0,0,0,0,0,0,0,0],
        [0,0,0,0,0,0,0,0,0],
        [0,0,0,0,0,0,0,0,0],
        [0,0,0,0,0,0,0,0,0]
    ]

r = getBestMove board1 1 4